from dataclasses import dataclass
import io
from typing import List
from googleapiclient.http import MediaFileUpload, MediaIoBaseUpload
from .gdrive_authenticator import GDriveAuthenticator


@dataclass
class GDriveWriter:
    output_folder_id: str

    def __post_init__(self):
        self.authenticator = GDriveAuthenticator()

    def write(self, relative_path: str, content: bytes) -> str:
        """
        Write the content to a file using Google Drive API.

        Args:
            relative_path: The relative path of the file to write. For example,
                if the output folder is "a", and the relative path is "b/c/d.txt",
                then this function will create a directory "b" in "a", and then
                create a directory "c" in "b", and then create a file "d.txt" in
                "c".
            content: The binary content to write to the file.
        Returns:
            The file ID of the file that was written.

        """
        service = self.authenticator.authenticate()
        parts = relative_path.split("/")
        dirs = parts[:-1]
        filename = parts[-1]
        parent_folder_id = self.makedirs(dirs)
        existing = self.find_files(parent_folder_id, filename)
        f = io.BytesIO(content)
        media = MediaIoBaseUpload(f, mimetype="application/octet-stream")
        if existing:
            print(f"File {relative_path} already exists. Overwriting...")
            existing_id = existing[0]["id"]
            file = (
                service.files().update(fileId=existing_id, media_body=media).execute()
            )
            return file["id"]
        else:
            file_metadata = {"name": filename, "parents": [parent_folder_id]}
            file = (
                service.files()
                .create(body=file_metadata, media_body=media, fields="id")
                .execute()
            )
            return file["id"]

    def makedirs(self, dirs: List[str]) -> str:
        """
        Create a nested subdirectory in the output folder. `dirs` represents
        one nested subdirectory. For example, if `dirs` is ["a", "b", "c"],
        then this function will create a directory "a" in the output folder,
        and then create a directory "b" in "a", and then create a directory
        "c" in "b". The function returns the ID of the last directory created.
        """
        service = self.authenticator.authenticate()
        parent_folder_id = self.output_folder_id
        for dir in dirs:
            existing = self.find_dirs(parent_folder_id, dir)
            if not existing:
                parent_folder_id = self._create_folder(service, parent_folder_id, dir)
            else:
                parent_folder_id = existing[0]["id"]
        return parent_folder_id

    def find_dirs(self, parent_folder_id: str, name: str) -> bool:
        """
        Check if a folder exists in the parent folder. Returns a list of
        folders with the same name in the parent folder.
        """
        service = self.authenticator.authenticate()
        response = (
            service.files()
            .list(
                q=f"'{parent_folder_id}' in parents and name='{name}' and mimeType='application/vnd.google-apps.folder'",
                spaces="drive",
                fields="files(id, name)",
            )
            .execute()
        )
        files = response.get("files", [])
        return files

    def find_files(self, parent_folder_id: str, name: str) -> bool:
        """
        Check if a file exists in the parent folder. Returns a list of
        files with the same name in the parent folder.
        """
        service = self.authenticator.authenticate()
        response = (
            service.files()
            .list(
                q=f"'{parent_folder_id}' in parents and name='{name}' and mimeType!='application/vnd.google-apps.folder'",
                spaces="drive",
                fields="files(id, name)",
            )
            .execute()
        )
        files = response.get("files", [])
        return files

    def _create_folder(self, service, parent_folder_id: str, folder_name: str) -> str:
        """
        Create a folder in the parent folder.

        Args:
            service: The Google Drive API service.
            parent_folder_id: The ID of the parent folder.
            folder_name: The name of the folder to create.
        Returns:
            The ID of the folder created.
        """
        file_metadata = {
            "name": folder_name,
            "parents": [parent_folder_id],
            "mimeType": "application/vnd.google-apps.folder",
        }
        file = service.files().create(body=file_metadata, fields="id").execute()
        return file.get("id")


if __name__ == "__main__":
    """
    python -m gdrive_watcher.gdrive_writer
    """
    from dotenv import load_dotenv
    import os

    load_dotenv()
    writer = GDriveWriter(output_folder_id=os.environ["OUTPUT_FOLDER_ID"])
    writer.write(relative_path="folder0/test.txt", content=b"test output 2")



from .gdrive_authenticator import GDriveAuthenticator


class GDriveReader:
    def __init__(self):
        self.authenticator = GDriveAuthenticator()

    def read_by_file_id(self, file_id: str) -> bytes:
        """
        Read the contents of a file using Google Drive API.
        """
        service = self.authenticator.authenticate()
        request = service.files().get_media(fileId=file_id)
        binary_content = request.execute()
        return binary_content



from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterable, List
import time
import google.auth
import google.auth.transport.requests
import google.auth.transport.urllib3
import googleapiclient.discovery
import googleapiclient.errors
from .gdrive_event import GDriveEvent, FileEventType


@dataclass
class GDrivePollWatcher:
    """
    Watch a Google Drive folder for changes.

    TODO: Add support for the user to set the time zone.
    """

    folder_id: str  # Google drive folder ID. If None, watch all folders.
    file_only: bool = False  # If True, only watch files, not folders.
    watch_start_time: datetime = datetime.now()
    sleep_time: int = 10  # Time (in seconds) to sleep between checks for changes
    # For authentication.
    scopes: List[str] = field(
        default_factory=lambda: ["https://www.googleapis.com/auth/drive"]
    )

    def watch(self) -> Iterable[GDriveEvent]:
        # Return newly added or modified files.

        # Ignore files modified before this time.

        service = self._authenticate()
        since = self.watch_start_time
        while True:
            recently_modified_files = list(
                self._get_recently_modified_files_recursively(
                    service, modified_since=since, folder_id=self.folder_id
                )
            )
            if self.file_only:
                recently_modified_files = [
                    file
                    for file in recently_modified_files
                    if not file["mimeType"].startswith(
                        "application/vnd.google-apps.folder"
                    )
                ]
            for file in recently_modified_files:
                yield GDriveEvent(
                    folder_id=self.folder_id,
                    file_id=file["id"],
                    file_name=file["name"],
                    relative_path=self._get_relative_path(
                        service=service,
                        parent_ids=file["parents"],
                        file_name=file["name"],
                    ),
                    event_type=FileEventType.UPDATED,
                    event_datetime=datetime.now(),
                    file_created_datetime=datetime.strptime(
                        file["createdTime"], "%Y-%m-%dT%H:%M:%S.%fZ"
                    ),
                    size=file.get("size"),
                    mime_type=file.get("mimeType"),
                )
            if recently_modified_files:
                modified_time_for_most_recent_file = datetime.strptime(
                    recently_modified_files[0]["modifiedTime"], "%Y-%m-%dT%H:%M:%S.%fZ"
                )
                since = modified_time_for_most_recent_file

            time.sleep(self.sleep_time)

    def _authenticate(self):
        creds, self.project_id = google.auth.default(scopes=self.scopes)
        self.service = googleapiclient.discovery.build("drive", "v3", credentials=creds)
        return self.service

    def _get_recently_modified_files_recursively(
        self, service, folder_id: str, modified_since: datetime
    ) -> Iterable[dict]:
        children = self._get_recently_modified_files(
            service,
            modified_since=modified_since,
            folder_id=folder_id,
            files_only=False,
        )
        for child in children:
            if child["mimeType"] == "application/vnd.google-apps.folder":
                yield from self._get_recently_modified_files_recursively(
                    service, folder_id=child["id"], modified_since=modified_since
                )
            else:
                yield child

    def _get_recently_modified_files(
        self, service, folder_id: str, modified_since: datetime, files_only=True
    ) -> List[dict]:
        """
        Example output:

        [
            {'size': '1024', 'id': '1LI8dE1zWjIAkHV2Qoekyz7HMhzli57Ir6sIi9OG8fks', 'name': 'test0', 'createdTime': '2022-12-27T04:21:43.801Z', 'modifiedTime': '2022-12-27T04:21:47.614Z'},
        ]
        """
        files = []
        try:
            page_token = None
            while True:
                response = (
                    service.files()
                    .list(
                        q=f"'{folder_id}' in parents",
                        spaces="drive",
                        fields="nextPageToken, "
                        "files(id, modifiedTime, name, size, parents, createdTime, mimeType)",
                        orderBy="modifiedTime desc",
                        pageToken=page_token,
                    )
                    .execute()
                )
                # print(response)
                reached_start_time = False
                for file in response.get("files", []):
                    # Save the file ID and modified time
                    modified_time = datetime.strptime(
                        file["modifiedTime"], "%Y-%m-%dT%H:%M:%S.%fZ"
                    )
                    if modified_time > modified_since:
                        if files_only:
                            if file["mimeType"] != "application/vnd.google-apps.folder":
                                assert "parents" in file, str(file)
                                files.append(file)
                        else:
                            files.append(file)
                    else:
                        reached_start_time = True
                        break
                if reached_start_time:
                    break
                page_token = response.get("nextPageToken", None)
                if page_token is None:
                    print(f"Found {len(files)} files in the folder")
                    break
                if len(files) == 0:
                    print("No files found.")
                    break

        except googleapiclient.errors.HttpError as error:
            print(f"An error occurred: {error}")
            raise
        return files

    def _get_relative_path(
        self, service, parent_ids: List[str], file_name: str
    ) -> List[str]:
        """
        Get the relative path of a file or folder.
        :param parent_ids: The IDs of the parent folders.
        :return: The relative path of the file or folder.
        """
        relative_path = []
        for parent_id in parent_ids:
            if parent_id == self.folder_id:
                continue
            parent = service.files().get(fileId=parent_id, fields="name").execute()
            relative_path.append(parent["name"])
        return relative_path + [file_name]



from dataclasses import dataclass, field
from typing import List
import google.auth
import google.auth.transport.requests
import googleapiclient.discovery


@dataclass
class GDriveAuthenticator:
    scopes: List[str] = field(
        default_factory=lambda: ["https://www.googleapis.com/auth/drive"]
    )

    def authenticate(self):
        if not hasattr(self, "service"):
            return self._authenticate()
        return self.service

    def _authenticate(self):
        creds, self.project_id = google.auth.default(scopes=self.scopes)
        self.service = googleapiclient.discovery.build("drive", "v3", credentials=creds)
        return self.service



if __name__ == "__main__":
    from datetime import datetime, timedelta
    from dotenv import load_dotenv
    import os
    from gdrive_watcher import GDriveWatcher
    from gdrive_watcher.gdrive_reader import GDriveReader

    load_dotenv()

    folder_id = os.environ["FOLDER_ID"]
    watcher = GDriveWatcher(
        folder_id=folder_id,
        watch_start_time=datetime.now() - timedelta(days=5),
    )
    reader = GDriveReader()
    for event in watcher.watch():
        print("-", event)
        if not event.is_folder:
            print("content:", reader.read_by_file_id(event.file_id))

    # TODO: try deleting a file manually, and see if the file still shows.



from dataclasses import dataclass
from datetime import datetime
import enum
from typing import List


class FileEventType(enum.Enum):
    """
    The type of event that occurred to a file.
    """

    CREATED = "created"
    UPDATED = "updated"


@dataclass
class GDriveEvent:
    """An event that occurred in the watched folder."""

    folder_id: str
    file_id: str
    file_name: str
    relative_path: List[str]  # Path relative to the watched folder
    event_type: FileEventType
    event_datetime: datetime = datetime.now()
    file_created_datetime: datetime = None
    size: int = None
    mime_type: str = None

    @property
    def is_folder(self):
        return self.mime_type == "application/vnd.google-apps.folder"



import os
import google.auth
import google.auth.transport.requests
import google.auth.transport.urllib3
import googleapiclient.discovery
import googleapiclient.errors
from dotenv import load_dotenv

# Load the environment variables
load_dotenv()

# Set up the Drive API client
SCOPES = ["https://www.googleapis.com/auth/drive"]
creds, project_id = google.auth.default(scopes=SCOPES)
print(f"Project ID: {project_id}")
service = googleapiclient.discovery.build("drive", "v3", credentials=creds)

# Set the ID of the folder you want to watch
folder_id = os.environ["FOLDER_ID"]

# List the files in the folder and save their metadata
def get_folder_contents():
    files = []
    try:
        page_token = None
        while True:
            response = (
                service.files()
                .list(
                    q=f"'{folder_id}' in parents",
                    spaces="drive",
                    fields="nextPageToken, "
                    "files(id, modifiedTime, name, parents, size, createdTime)",
                    orderBy="modifiedTime desc",
                    pageToken=page_token,
                )
                .execute()
            )
            print(response)
            for file in response.get("files", []):
                # Save the file ID and modified time
                files.append({"id": file["id"], "modifiedTime": file["modifiedTime"]})
            page_token = response.get("nextPageToken", None)
            if page_token is None:
                print(f"Found {len(files)} files in the folder")
                break
            if len(files) == 0:
                print("No files found.")
                break

    except googleapiclient.errors.HttpError as error:
        print(f"An error occurred: {error}")
        raise
    return files


# Save the initial list of files
previous_files = get_folder_contents()

# Set up a loop to check for changes every 60 seconds
while True:
    # Get the current list of files
    current_files = get_folder_contents()

    # Compare the lists and print a message for any new or modified files
    for file in current_files:
        print(f'Checking file: {file["id"]}')
        if file not in previous_files:
            print(f'------ New file: {file["id"]}')

    break
